import { SwaggerParser } from './SwaggerParser';
import { isNil } from 'lodash';
import { ConduitRouteActions, Indexable } from '@conduitplatform/grpc-sdk';
import { SwaggerRouterMetadata } from '../types';
import { ConduitRoute } from '../classes';

export class SwaggerGenerator {
  private readonly _swaggerDoc: Indexable;
  private readonly _routerMetadata: SwaggerRouterMetadata;
  private readonly _stringifiedGlobalSecurityHeaders: string;
  private _parser: SwaggerParser;

  constructor(routerMetadata: SwaggerRouterMetadata) {
    this._swaggerDoc = {
      openapi: '3.0.0',
      info: {
        version: '1.0.0',
        title: 'Conduit',
      },
      paths: {},
      components: {
        schemas: {
          ModelId: {
            type: 'string',
            format: 'uuid',
          },
        },
        securitySchemes: routerMetadata.securitySchemes,
      },
    };
    this._parser = new SwaggerParser();
    this._routerMetadata = routerMetadata;
    this._stringifiedGlobalSecurityHeaders = JSON.stringify(
      this._routerMetadata.globalSecurityHeaders,
    );
  }

  get swaggerDoc() {
    return this._swaggerDoc;
  }

  addRouteSwaggerDocumentation(route: ConduitRoute) {
    const method = this._extractMethod(route.input.action);
    const baseName = route.input.path.toString().replace('/hook', '').slice(1);
    const prefix =
      baseName.indexOf('/') !== -1 ? baseName.substring(0, baseName.indexOf('/')) : '';
    let serviceName: string;
    if (baseName.includes('admin')) {
      serviceName = 'admin';
    } else if (prefix.trim() === '') {
      serviceName = 'core';
    } else {
      serviceName = prefix;
    }
    const routeDoc: Indexable = {
      summary: route.input.name,
      description: route.input.description,
      tags: [serviceName],
      parameters: [],
      responses: {
        200: {
          description: 'Successful Operation',
          content: {
            'application/json': {
              schema: {},
            },
          },
        },
      },
      security: JSON.parse(this._stringifiedGlobalSecurityHeaders),
    };

    if (
      !isNil(route.input.urlParams) &&
      (route.input.urlParams as unknown as string) !== ''
    ) {
      for (const name in route.input.urlParams) {
        routeDoc.parameters.push({
          name,
          in: 'path',
          required: true,
          schema: this._parser.extractTypes('url', route.input.urlParams, true),
        });
      }
    }

    if (
      !isNil(route.input.queryParams) &&
      (route.input.queryParams as unknown as string) !== ''
    ) {
      for (const name in route.input.queryParams) {
        routeDoc.parameters.push({
          name,
          in: 'query',
          schema: this._parser.extractTypes('query', route.input.queryParams, true),
        });
      }
    }

    if (
      !isNil(route.input.bodyParams) &&
      (route.input.bodyParams as unknown as string) !== ''
    ) {
      routeDoc['requestBody'] = {
        description: route.input.description,
        content: {
          'application/json': {
            schema: this._parser.extractTypes('body', route.input.bodyParams, true),
          },
        },
        required: true,
      };
    }

    this._routerMetadata.setExtraRouteHeaders(route, routeDoc);

    const returnDefinition = this._parser.extractTypes(
      route.returnTypeName,
      route.returnTypeFields,
      false,
    );
    routeDoc.responses[200].content['application/json'].schema = {
      $ref: `#/components/schemas/${route.returnTypeName}`,
    };
    if (!this._swaggerDoc.components['schemas'][route.returnTypeName]) {
      this._swaggerDoc.components['schemas'][route.returnTypeName] = returnDefinition;
    }
    const path =
      this._routerMetadata.urlPrefix + route.input.path.replace(/(:)(\w+)/g, '{$2}');
    if (this._swaggerDoc.paths.hasOwnProperty(path)) {
      this._swaggerDoc.paths[path][method] = routeDoc;
    } else {
      this._swaggerDoc.paths[path] = {};
      this._swaggerDoc.paths[path][method] = routeDoc;
    }
  }

  private _extractMethod(action: string) {
    switch (action) {
      case ConduitRouteActions.GET: {
        return 'get';
      }
      case ConduitRouteActions.POST: {
        return 'post';
      }
      case ConduitRouteActions.DELETE: {
        return 'delete';
      }
      case ConduitRouteActions.UPDATE: {
        return 'put';
      }
      case ConduitRouteActions.PATCH: {
        return 'patch';
      }
      default: {
        return 'get';
      }
    }
  }
}
